/*
 *  acmi: A program to compute Air Core Mutual Inductances.
 *
 *      origin: http://www.abelian.demon.co.uk/acmi/
 *  maintainer: Paul Nicholson, paul@abelian.demon.co.uk
 *
 *  Thanks go to Mark Rzeszotarski for suggesting the use of Grover's tables,
 *  and to Antonio Carlos M. de Queiroz for pointing out a serious bug
 *  in v0.1c.
 * 
 *  Many thanks to Terry Fritz, <twftesla@uswest.net>,
 *              Barton B. Anderson, <tesla123@pacbell.net>,
 *          and Bert Hickman, <bert.hickman@aquila.net>, 
 *  for their invaluable help in providing validation measurements.
 */

#define VERSION "v0.9a, Jul 29th 2003"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

/*
 *  Flags set by command line options.
 */

int opt_HEADING = 1;             /* Non-zero if column headings are required */
int opt_HELP = 0;                         /* Non-zero if usage help required */
int opt_HTML = 0;                                    /* Generate HTML output */
int opt_CSV = 0;                /* Generate Comma Separated Variables output */
int opt_SSV = 0;                /* Generate Space Separated Variables output */
int opt_M = 1;                    /* Non-zero if mutual inductances required */
int opt_L = 1;                      /* Non-zero if self inductances required */
int opt_ALL = 0;                       /* Non-zero for a more verbose output */

char *unit_str = "m";             /* Units indicator for output measurements */
double unit_scale = 1.0;           /* Multiplier applied to input measurements,
                                        and divider for output measurements  */

double *S;                             /* Special variable for coil profiles */

#define PREC (1 - 1e-9)
#define MU (4 * 3.14159265 * 1.0E-7)

double fmaxr = -1e99;
double fminr = +1e99;
double fmaxz = -1e99;
double fminz = +1e99;

struct NODE *field_maxr = NULL, *field_minz = NULL, *field_maxz = NULL;

/*****************************************************************************/
/*  Miscellaneous Stuff                                                      */
/*****************************************************************************/

void bailout( char *format, ...)
{
   char temp[100];

   vsprintf( temp, format, &format + 1);
   fprintf( stderr, "ERROR: %s", temp);
   exit( 1);
}

int is_vowel( char c)
{
   return c == 'A' || c == 'E' || c == 'I' ||
          c == 'O' || c == 'U' || c == '_';
}

void abbreviate( char *abbr, char *str)
{
   char *d = abbr, *s = str;

   if( strlen( str) < 4) { strcpy( abbr, str); return; }

   while( *s) if( (d-abbr < 3 && !is_vowel(*s)) ||
                  isdigit( *s) || s == str + strlen(str) - 1) *d++ = *s++;
              else s++;

   *d = 0;
}

void *safe_malloc( int size)
{
   void *p = malloc( size);  if( !p) bailout( "not enough memory");
   memset( p, 0, size);
   return p;
}

double awg_to_m( double a)
{
   return 7.348e-3/pow(1.122932, a-1)/2;
}

/*****************************************************************************/
/*  Grover's Tables                                                          */
/*****************************************************************************/

/*
 *  Tables from F.W. Grover, 'Inductance Calculations', 1946;
 *  Reprinted by Dover, 1962.
 */

double grover13[] = 
{
   0.008297,  /* 0.1 */
   0.007810, 0.007371, 0.006974, 0.006611, 0.006278, 
   0.005970, 0.005685, 0.005420, 0.005173,
   0.004941, /* 0.2 */
   0.004723, 0.004518, 0.004325, 0.004142, 0.003969,
   0.003805, 0.003649, 0.003500, 0.003359,
   0.003224, /* 0.3 */
   0.003095, 0.002971, 0.002853, 0.002740, 0.0026317,
   0.0025276, 0.0024276, 0.0023315, 0.0022391, 
   0.0021502, /* 0.4 */
   0.0020646, 0.0019821, 0.0019026, 0.0018259, 0.0017519,
   0.0016805, 0.0016116, 0.0015451, 0.0014808, 
   0.0014186, /* 0.5 */
   0.0013585, 0.0013004, 0.0012443, 0.0011900, 0.0011374,
   0.0010865, 0.0010373, 0.0009897, 0.0009436,
   0.0008990, /* 0.6 */
   0.0008558, 0.0008141, 0.0007736, 0.0007345, 0.0006966,
   0.0006600, 0.0006246, 0.0005903, 0.0005571,
   0.0005251, /* 0.7 */
   0.0004941, 0.0004642, 0.0004353, 0.0004074, 0.0003805,
   0.0003545, 0.0003295, 0.0003054, 0.0002823,
   0.00025998, /* 0.8 */
   0.00023859, 0.00021806, 0.00019840, 0.00017959, 0.00016162,
   0.00014450, 0.00012821, 0.00011276, 0.00009815,
   0.00008438  /* 0.9 */
};

double grover14[] = 
{
   0.079093, /* -6.0 */
   0.077647, 0.076200, 0.074753, 0.073306, 0.071860,
   0.070413, 0.068966, 0.067520, 0.066073,
   0.064626, /* -5.0 */
   0.063180, 0.061733, 0.060287, 0.058840, 0.057394,
   0.055947, 0.054500, 0.053055, 0.051609,
   0.050163, /* -4.0 */
   0.048717, 0.047272, 0.045827, 0.044382, 0.042938,
   0.041494, 0.040051, 0.038608, 0.037167,
   0.035727, /* -3.0 */
   0.034288, 0.032851, 0.031416, 0.029984, 0.028554, 
   0.027128, 0.025707, 0.024291, 0.022881,
   0.021478, /* -2.0 */
   0.020084, 0.018700, 0.017329, 0.015972, 0.014632,
   0.013311, 0.012013, 0.010742, 0.009502,
   0.008297  /* -1.0 */
};

double grover15[] = 
{
   .39227 - 9, /* -4.0 */
   .54228 - 9, .69229 - 9, .85230 - 9, .99232 - 9, .14234 - 8,
   .29237 - 8, .44240 - 8, .59244 - 8, .74250 - 8,
   .89257 - 8, /* -3.0 */
   .04265 - 7, .19276 - 7, .34289 - 7, .49306 - 7, .64327 - 7,
   .79354 - 7, .94388 - 7, .09430 - 6, .22484 - 6, 
   .39551 - 6, /* -2.0 */
   .54636 - 6, .69744 - 6, .84879 - 6, .00051 - 5, .15268 - 5,
   .30542 - 5, .45891 - 5, .61334 - 5, .76899 - 5,
   .92622 - 5  /* -1.0 */
};

/*
 *  Compute the mutual inductance between two circular filaments, perpendicular
 *  distance d apart, with radii r1 and r2. Our variable K stands for Grover's
 *  k' squared.
 */

double filament( double r1, double r2, double d, double w)
{
   double K, fk, g1, g2;

   K = ((r1 - r2)*(r1 - r2) + d*d) / ((r1+r2)*(r1+r2) + d*d);

   if( K < 1e-8)
   {  
      return MU * r1 * (log(8*r1/w) - 2) + (MU * r1/4.0);
   }

   if( K < 0.1)
   {
      int i = (int)(log10(K) * 10);

      if( i < -60) fk = 0.014468 * ( log10(1/K)-0.53307);
      else
      {
         g1 = grover14[ i+60];   g2 = grover14[ i+60-1];
         fk = ((log10(K)-(i/10.0))/-0.1) * (g2-g1) + g1;
      }
   }
   else
   if( K >= 0.9)
   {
      int i = (int)(log10(1-K) * 10);
      if( i < -40) fk = pow( 10.0, .39224 - 3.0 + 1.5 * log10(1-K));
      else
      {
         g1 = grover15[ i+40];   g2 = grover15[ i+40-1];
         fk = pow( 10.0, ((log10(1-K)-(i/10.0))/-0.1) * (g2-g1) + g1);
      }
   }
   else
   {
      int i = (int)(K * 100);
      g1 = grover13[ i-10];   g2 = grover13[ i-10+1];
      fk = (K-(i/100.0))/0.01 * (g2-g1) + g1;
   } 

   /*
    *  The factor 100.0 converts from Grover's cgs to SI, and 1e-6 converts
    *  from uH to H.
    */

   return 1e-6 * 100.0 * sqrt(r1 * r2) * fk;
}

/*****************************************************************************/
/*  Winding Descriptions                                                     */
/*****************************************************************************/

/*
 *  We use an instance of COIL structure for each winding. The physical
 *  dimensions describing the winding are stored as expressions, each in the
 *  form of a linked list of NODE structures.
 */

struct COIL
{
   char name[100];
   char abbr[10];
   struct NODE * radius1;                               /* Radius at one end */
   struct NODE * radius2;                             /* Radius at other end */
   struct NODE * height1;                               /* Height of one end */
   struct NODE * height2;                             /* Height of other end */
   struct NODE * wirad;                                       /* Wire radius */
   struct NODE * turns;                                   /* Number of turns */
   struct NODE * current;                            /* Current distribution */
   double R;                                     /* DC resistance of winding */
   double L;                                                /* DC inductance */
   struct COIL *next;
}
 *coils = 0, **coil_ep = &coils;        /* A linked list with an end pointer */

/*
 *  Allocate a COIL structure, initialise it, and tack it onto the end of a 
 *  linked list.
 */

struct COIL *alloc_coil( char *name)
{
   struct COIL *c = *coil_ep = safe_malloc( sizeof( struct COIL));

   *(coil_ep = &c->next) = 0;
 
   strcpy( c->name, name);
   return c;
}

/*
 *  Lookup a coil by name, and return its pointer or null.
 */

struct COIL *find_coil( char *name)
{
   struct COIL *c = coils;

   while( c && strcmp( name, c->name)) c = c->next;
   return c;
}

/*****************************************************************************/
/*  Expression Evaluation Stuff                                              */
/*****************************************************************************/

/* 
 *  An expression is a linked list of nodes in postfix order.  Each node is one
 *  of the following types.
 */

#define NT_CONST 0                                     /* A numeric constant */
#define NT_OPER2 1                          /* An binary arithmetic operator */
#define NT_OPER3 2                          /* An binary arithmetic operator */
#define NT_DEFN  3                                       /* A defined symbol */
#define NT_COIL  4                                       /* A coil reference */
#define NT_RANGE 5                                       /* A range variable */
#define NT_AFCN1 6          /* Call an arithmetic function with one argument */
#define NT_CFCN1 7                 /* Call a coil function with one argument */
#define NT_CFCN2 8                /* Call a coil function with two arguments */

/*
 *  The above types describe the expression, the following types carry meta
 *  information.
 */ 

#define NT_UNIT  9                                    /* A units description */

struct NODE 
{
   int type;                             /* One of the above type indicators */
   union 
   {
      double val;
      int op;
      struct DEFN *defn;
      struct RANGE *range;
      double (*func)();
      struct UNIT *unit;
      struct COIL *coil;
   } u; 

   struct NODE *next;                                 /* Linked list pointer */
};

struct NODE *ambient = NULL;   /* Ambient temperature - used for resistances */

struct NODE *alloc_node( int type)
{
   struct NODE *n = safe_malloc( sizeof( struct NODE));

   n->type = type;
   n->next = NULL;
   return n;
}

/*
 *  Return a pointer to the last node of an expression.
 */

struct NODE *end_of( struct NODE *n)
{
   while( n->next) n = n->next;
   return n;
}

/*
 *  A variable definition binds a symbol name to its defining expression.
 */

struct DEFN
{
   char name[100];                         /* Symbol name of this definition */
   struct NODE *expr;                                      /* Its expression */
   struct DEFN *next;                                 /* Linked list pointer */
}
 *defns, **defn_ep = &defns;                 /* A linked list of definitions */

struct DEFN *find_defn( char *name)
{
   struct DEFN *d = defns;

   for( ; d && strcmp( d->name, name); d = d->next);
   return d;
}

struct DEFN *alloc_defn( char *name)
{
   struct DEFN *d = *defn_ep = safe_malloc( sizeof( struct DEFN));

   *(defn_ep = &d->next) = 0;
   strcpy( d->name, name);

   return d;
}

/*
 *  A structure to record a range variable definition.
 */

struct RANGE 
{
   struct NODE *from;                                       /* Initial value */
   struct NODE *to;                                           /* Final value */
   struct NODE *step;                                      /* Post-increment */
   double val;                                              /* Current value */
   double opt;                                              /* Optimum value */
   struct RANGE *next;
}
 *ranges = 0, **range_ep = &ranges;

struct RANGE *alloc_range( void)
{
   struct RANGE *r = safe_malloc( sizeof( struct RANGE));

   *range_ep = r;
   *(range_ep = &r->next) = 0;

   return r;
}

void save_optimum( void)
{
   struct RANGE *r = ranges;
   for(; r; r = r->next) r->opt = r->val;
}

void restore_optimum( void)
{
   struct RANGE *r = ranges;
   for(; r; r = r->next) r->val = r->opt;
}

/*
 *  To handle units conversions, we jam a special node into the expression,
 *  in order to hold conversion data.
 */

struct UNIT
{
   double factor;                         /* Factor for conversion to metric */
   double (*cvt)(double);                     /* Conversion function, if any */
   char *symbol;                                            /* Symbol string */
};

/*
 *  Take an unscaled postfix expression and prefix it with a units conversion.
 */

struct NODE *scale_expression( struct NODE *n, 
                               double factor, double (*cvt)(), char *symbol)
{
   struct NODE *t = alloc_node( NT_UNIT);
   t->u.unit = safe_malloc( sizeof( struct UNIT));
   t->u.unit->factor = factor;
   t->u.unit->cvt = cvt;
   t->u.unit->symbol = symbol;
   t->next = n;
   return t;
}

/*
 *  Extract the units conversion for the given expression.
 */

struct UNIT *extract_units( struct NODE *n)
{
   for( ; n; n = n->next) if( n->type == NT_UNIT) return n->u.unit;
   return 0;
}

/*
 *  A stack of doubles for expression evaluation, along with a stack pointer.
 */

double estack[1000], *estack_p = estack;

/*
 *  A stack of coil pointers for coil functions.
 */

struct COIL *cstack[10], **cstack_p = cstack;

/*
 *  A recursive function to evaluate the given expression.
 */

double evaluate( struct NODE *n)
{
   double d1, d2, d3;
   struct COIL *c1, *c2;

   for( ; n; n = n->next)
      switch( n->type)
      {
         case NT_CONST: /* 
                         *  Just push a constant value onto the stack.
                         */

                        *estack_p++ = n->u.val; break;

         case NT_OPER2: /*
                         *  Pop the top two values off the stack, apply the
                         *  appropriate operator to them, and push the result
                         *  back onto the stack.
                         */
 
                        d2 = *--estack_p;
                        d1 = *--estack_p;
               
                        switch( n->u.op)
                        {
                           case '+':  *estack_p++ = d1 + d2; break;
                           case '-':  *estack_p++ = d1 - d2; break;
                           case '*':  *estack_p++ = d1 * d2; break;
                           case '/':  *estack_p++ = d1 / d2; break;
                           case '>':  *estack_p++ = d1 > d2; break;
                           case '<':  *estack_p++ = d1 < d2; break;
                           case '=':  *estack_p++ = d1 == d2; break;
                           case '%':  *estack_p++ = ((int)d1) % ((int)d2);
                                                    break;
                        }
                        break;  
               
         case NT_OPER3: /*
                         *  Pop the three values off the stack, and push the
                         *  selected result back on.
                         */
 
                        d3 = *--estack_p;
                        d2 = *--estack_p;
                        d1 = *--estack_p;
                        *estack_p++ = d1 ? d2 : d3;
                        break; 
               
         case NT_DEFN:  /*
                         *  Evaluate the defining expression for the symbol
                         *  and push the result onto the stack.
                         */

                        *estack_p++ = evaluate( n->u.defn->expr);  break;

         case NT_RANGE: /*
                         *  Push the current value of the range variable onto
                         *  the stack.
                         */

                        *estack_p++ = n->u.range->val;  break;

         case NT_AFCN1: /*
                         *  Call an arithmetic function.
                         */

                        d1 = *--estack_p; *estack_p++ = n->u.func(d1);  break;

         case NT_COIL:  /*
                         *  Push a coil reference onto the coil stack.
                         */

                        if( cstack_p - cstack > 2) bailout( "cstack ovfl");
                        *cstack_p++ = n->u.coil; break; 

                        /*
                         *  Coil functions, of one or two arguments.
                         */

         case NT_CFCN1: c1 = *--cstack_p; *estack_p++ = n->u.func(c1);  
                        break;
         case NT_CFCN2: c2 = *--cstack_p; 
                        c1 = *--cstack_p; *estack_p++ = n->u.func(c1, c2);
                        break;
      }

   /*
    *  By now the result is the top item on the stack.
    */

   return *--estack_p;
}

/*
 *  Evaluate an expression, as above, but then apply the units conversion, if
 *  any.
 */

double evaluate_scaled( struct NODE *n)
{
   struct UNIT *unit = extract_units( n);

   if( !unit) return evaluate( n);

   if( unit->cvt) return unit->cvt( evaluate( n));
   else return unit->factor * evaluate( n);
}

/*
 *  Return zero if the given expression, or any of its sub-expressions
 *  references a range variable. Otherwise, return 1 to indicate that the
 *  expression has a fixed value.
 */

int fixed( struct NODE *n)
{
   for( ; n; n = n->next)
      switch( n->type)
      {
         case NT_DEFN: if( !fixed( n->u.defn->expr)) return 0;
                       break;
         case NT_RANGE: return 0;
      }
   return 1;
}

/*
 *  Return 1 if the given expression, or any of its sub-expressions,
 *  depends on any results from a specific coil.
 */

int depends_this_coil( struct NODE *n, struct COIL *c)
{
   for( ; n; n = n->next)
      switch( n->type)
      {
          case NT_DEFN: if( depends_this_coil( n->u.defn->expr, c)) return 1;
                        break;
         case NT_RANGE: if( depends_this_coil( n->u.range->from, c) ||
                            depends_this_coil( n->u.range->to, c) ||
                            depends_this_coil( n->u.range->step, c)) return 1; 
                        break;
          case NT_COIL: if( n->u.coil == c) return 1;
                        break;
      }

   return 0;
}

/*
 *  Return 1 if the given expression, or any of its sub-expressions, 
 *  refers to the special variable S.
 */

int depends_S( struct NODE *n)
{
   for( ; n; n = n->next)
      switch( n->type)
      {
         case NT_DEFN: if( depends_S( n->u.defn->expr)) return 1;
                       break;
         case NT_RANGE: if( depends_S( n->u.range->from) ||
                            depends_S( n->u.range->to) ||
                            depends_S( n->u.range->step)) return 1; 
                       break;
         case NT_CONST: if( &n->u.val == S) return 1;
                       break;
      }

   return 0;
}

/*****************************************************************************/
/*  Output Descriptions                                                      */
/*****************************************************************************/

/*
 *  Each output column is described by an instance of struct SHOW.
 */

struct SHOW
{
   char heading[100];
   char format[50];
   int len;                                       /* The width of the column */
   struct NODE *expr;      /* The expression which generates the column data */
   struct SHOW *next;
   void (*fcn)( int, double);   
}
 *shows = 0, **shows_ep = &shows;          /* A linked list with end pointer */

struct SHOW *alloc_show( struct NODE *n)
{
   struct SHOW *s = *shows_ep = safe_malloc( sizeof( struct SHOW));

   *(shows_ep = &s->next) = 0;
   s->expr = n;
   return s;
}

/*****************************************************************************/
/*  Input File Parser                                                        */
/*****************************************************************************/

/*
 *  A simple top-down recursive descent parser for the input file.
 */

char pbuff[ 500], *pb;           /* Input line buffer, and a pointer into it */
int lino;                                             /* Current line number */
int eof_flag = 0;                       /* Non-zero when end-of-file reached */

/*
 *  Refill the parse buffer with the next line of input.
 */

int parse_reload( void)
{
   *(pb = pbuff) = 0;

   if( fgets( pbuff, 499, stdin)) { lino++; return 1; }

   eof_flag = 1;
   return 0;
}

/*
 *  Report an error in the input file, and then exit.
 */

void parse_error( char *msg, ...)
{
   char temp[100];

   vsprintf( temp, msg, (char **)(&msg + 1));
   bailout( "input file, line %d: %s\n", lino, temp);
}

/*
 *  Discard white space, blank lines, and comments.
 */

void skip_whitespace( void)
{
   while( 1)
      if( !*pb)                                          /* Skip blank lines */
      {
         if( !parse_reload()) return;
      }
      else
      if( isspace( *pb)) pb++;                            /* Skip whitespace */
      else
      if( *pb == ';') *pb = 0;                              /* Skip comments */
      else
         return;
}

int is_separator( char c)
{
   return !c || isspace(c) ||
           c == '{' || c == '}' || c == '[' || c == ']' || c == '(' || c == ')';
}

/*
 *  Parse a variable name: A letter followed by a string of letters, digits,
 *  and underscores.  We translate everything to upper case internally.
 */

int parse_name( char *s)
{
   int n = 0;

   skip_whitespace();

   if( isalpha(*pb))
      while( *pb && (isalnum( *pb) || *pb == '_')) *s++ = toupper( *pb++), n++;

   *s = 0;
   return n;
}

/*
 *  Test for a character symbol.
 */

int parse_char( char c)
{
   skip_whitespace();
   if( *pb == c) { pb++; return c; }
   return 0;
}

int parse_qstring( char *s)
{
   int n = 0;
   if( !parse_char( '"')) return 0;

   while( *pb && *pb != '"') *s++ = *pb++, n++;
   if( !*pb) parse_error( "unterminated string");
   parse_char( '"');
   return n;
}

/* 
 *  Test for a keyword symbol.
 */

int parse_keyword( char *target)
{
   char *p;
   int m, n;

   skip_whitespace();

   for( p=pb, m = n = strlen( target); n--; p++, target++)
       if( toupper( *p) != toupper( *target)) return 0;
  
   if( !is_separator( *p)) return 0;
 
   pb += m;
   return 1;
}

struct NODE * parse_coil_reference( void)
{
   char name[100];
   struct NODE *n = alloc_node( NT_COIL);

   if( !parse_name( name)) parse_error( "missing coil name");
   if( (n->u.coil = find_coil( name)) == NULL) parse_error( "undefined coil");
   return n;
}

double compute_self_inductance( struct COIL *);
double compute_resistance( struct COIL *c);
double compute_length( struct COIL *c);
double compute_mutual_inductance( struct COIL *ca, struct COIL *cb);
double compute_coupling_coeff( struct COIL *ca, struct COIL *cb);

struct NODE * parse_coil_function( char *name)
{
   struct NODE *n = 0; 

   if( !strcmp( name, "SELF") && parse_char( '(')) 
      (end_of(n = parse_coil_reference())->next =
                 alloc_node( NT_CFCN1))->u.func = compute_self_inductance;
   else
   if( !strcmp( name, "RESISTANCE") && parse_char( '(')) 
      (end_of(n = parse_coil_reference())->next =
                 alloc_node( NT_CFCN1))->u.func = compute_resistance;
   else
   if( !strcmp( name, "LENGTH") && parse_char( '(')) 
      (end_of(n = parse_coil_reference())->next =
                 alloc_node( NT_CFCN1))->u.func = compute_length;
   else
   if( !strcmp( name, "MUTUAL") && parse_char( '(')) 
   {
      n = parse_coil_reference();
      if( !parse_char( ',')) parse_error( "faulty arguments to mutual()");

      (end_of( end_of(n)->next = parse_coil_reference())->next =
                 alloc_node( NT_CFCN2))->u.func = compute_mutual_inductance;
   }
   else
   if( !strcmp( name, "COUPLING") && parse_char( '(')) 
   {
      n = parse_coil_reference();
      if( !parse_char( ',')) parse_error( "faulty arguments to coupling()");

      (end_of( end_of(n)->next = parse_coil_reference())->next =
                 alloc_node( NT_CFCN2))->u.func = compute_coupling_coeff;
   }
   
   return n;
}

struct NODE * parse_function( char *name)
{
   struct NODE *n = 0, *parse_tertiary_expression( void);

   if( !strcmp( name, "SQRT") && parse_char( '(')) 
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = sqrt;
   else
   if( !strcmp( name, "SIN") && parse_char( '(')) 
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = sin;
   else
   if( !strcmp( name, "COS") && parse_char( '(')) 
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = cos;
   else
   if( !strcmp( name, "EXP") && parse_char( '(')) 
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = exp;
   else
   if( !strcmp( name, "LN") && parse_char( '(')) 
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = log;
   else
   if( !strcmp( name, "ABS") && parse_char( '(')) 
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = fabs;
   else
   if( !strcmp( name, "INT") && parse_char( '(')) 
      (end_of(n = parse_tertiary_expression())->next =
                 alloc_node( NT_AFCN1))->u.func = floor;
   else
   if( (n = parse_coil_function( name)) == 0) return 0;

   if( !parse_char( ')')) parse_error( "missing closing ')'");
   return n;
}

/*
 *  Parse a value: this is either a bracketed sub-expression, a defined
 *  variable, a range definition, a function call, or a numeric constant. 
 */

struct NODE * parse_value( void)
{
   char *s, name[100];
   struct NODE *n, *parse_tertiary_expression( void);

   if( parse_char( '(')) 
   {
      n = parse_tertiary_expression();
      if( !parse_char( ')')) parse_error( "missing )");
   }   
   else
   if( parse_name( name))
   {
      if( (n = parse_function( name)) == 0 &&
          ((n = alloc_node( NT_DEFN))->u.defn = find_defn(name)) == 0)
             parse_error( "undefined symbol [%s]", name);
   } 
   else
   if( parse_char( '['))
   {
      (n = alloc_node( NT_RANGE))->u.range = alloc_range();

      if( !parse_keyword( "from")) parse_error( "missing 'from' in range");
      n->u.range->from = parse_tertiary_expression();
      if( !parse_keyword( "to")) parse_error( "missing 'to' in range");
      n->u.range->to = parse_tertiary_expression();
      if( !parse_keyword( "step")) parse_error( "missing 'step' in range");
      n->u.range->step = parse_tertiary_expression();

      if( !parse_char( ']')) parse_error( "range missing closing ']'");
   }
   else
   {
      skip_whitespace(); s = pb;
      (n = alloc_node( NT_CONST))->u.val = strtod( pb, &pb);
      if( s == pb) parse_error( "not a valid expression");
   }

   return n;
}

/*
 *  Parse high-precedence operators.
 */

struct NODE * parse_mul_expression( void)
{
   int c;
   struct NODE *n = parse_value();

   while( (c = parse_char( '*')) != 0 ||
          (c = parse_char( '/')) != 0 ||
          (c = parse_char( '%')) != 0)
   {
      struct NODE *e = end_of(n)->next = parse_value();
      (end_of(e)->next = alloc_node( NT_OPER2))->u.op = c;
   }
   
   return n;
}

/*
 *  Parse low-precedence operators.
 */

struct NODE * parse_sum_expression( void)
{
   int c;
   struct NODE *n = parse_mul_expression();

   while( (c = parse_char( '+')) != 0 ||
          (c = parse_char( '-')) != 0) 
   {
      struct NODE *e = end_of(n)->next = parse_mul_expression();
      (end_of(e)->next = alloc_node( NT_OPER2))->u.op = c;
   }

   return n;
}

/*
 *  Parse very low-precedence operators.
 */

struct NODE * parse_test_expression( void)
{
   int c;
   struct NODE *n = parse_sum_expression();

   while( (c = parse_char( '>')) != 0 ||
          (c = parse_char( '<')) != 0 ||
          (c = parse_char( '=')) != 0) 
   {
      struct NODE *e = end_of(n)->next = parse_test_expression();
      (end_of(e)->next = alloc_node( NT_OPER2))->u.op = c;
   }

   return n;
}

struct NODE * parse_tertiary_expression( void)
{
   struct NODE *n = parse_test_expression();

   if( !parse_char( '?')) return n;

   end_of( n)->next = parse_tertiary_expression();

   if( !parse_char( ':')) parse_error( "expecting ':' in tertiary operator"); 
   end_of( n)->next = parse_tertiary_expression();

   end_of( n)->next = alloc_node( NT_OPER3);
   return n;
}

/*
 *  Parse a variable definition:  name = expression
 */

int parse_defn( char *name)
{
   if( !parse_char( '=')) return 0;                      /* Not a definition */

   if( find_defn( name)) bailout( "duplicate definition of %s", name);

   alloc_defn( name)->expr = parse_tertiary_expression();
   return 1;
}

/*
 *  Parse an expression followed by an optional units symbol.
 */

struct NODE *parse_scaled_expression( void)
{
   struct NODE *n = parse_tertiary_expression();

   if( parse_char( '"') ||
       parse_keyword( "inches") ||
       parse_keyword( "inch")) return scale_expression( n, 1/39.37, 0, "\"");

   if( parse_keyword( "mm")) return scale_expression( n, 1e-3, 0, "mm");

   if( parse_char( '\'')) return scale_expression( n, 12/39.37, 0, "'");

   if( parse_keyword( "awg")) return scale_expression( n, 0, awg_to_m, "awg");

   return scale_expression( n, unit_scale, 0, unit_str);
}

/*
 *  Parse a coil description clause: { parameters ... }
 */

int parse_coil( char *name)
{
   struct COIL *c;

   if( !parse_char( '{')) return 0;                /* Not a coil description */

   /*
    *  Allocate a coil description structure, save its name and abbreviated
    *  name.
    */

   c = alloc_coil( name);
   abbreviate( c->abbr, c->name);

   /*
    *  Parse all the coil parameters.
    */

   while( !parse_char( '}')) 
   {
      if( eof_flag) parse_error( "missing closing }");

      if( parse_keyword( "radius1")) 
         c->radius1 = parse_scaled_expression();
      else
      if( parse_keyword( "radius2"))
         c->radius2 = parse_scaled_expression();
      else
      if( parse_keyword( "radius")) 
         c->radius1 =  c->radius2 = parse_scaled_expression(); 
      else
      if( parse_keyword( "height1"))
         c->height1 = parse_scaled_expression();
      else
      if( parse_keyword( "height2"))
         c->height2 = parse_scaled_expression();
      else
      if( parse_keyword( "height"))
      {
         c->height1 = c->height2 = parse_scaled_expression();
      }
      else
      if( parse_keyword( "conductor"))
      {
         c->wirad = parse_scaled_expression();
         if( depends_S( c->wirad))
            parse_error( "Paul was too lazy to code for non-uniform wire");
      }
      else
      if( parse_keyword( "turns")) 
      {
         c->turns = parse_tertiary_expression();
         if( depends_S( c->turns))
            parse_error( "Not meaningful for turns to depend on S");
      }
      else
      if( parse_keyword( "current"))
         c->current = parse_tertiary_expression();
      else
         parse_error( "unrecognised coil parameter");
   }

   /*
    *  Make sure we've been given a full set of parameters.
    */

   if( !c->radius1 || !c->radius2) parse_error( "missing radius in coil");
   if( !c->height1 || !c->height2) parse_error( "missing height in coil");
   if( !c->turns) parse_error( "missing turns in coil");
   if( !c->wirad) parse_error( "missing conductor in coil");

   if( c->radius1 != c->radius2 && 
       (depends_S( c->radius1) || depends_S( c->radius2)))
      parse_error( "requires single radius expression for S dependancy");

   if( c->height1 != c->height2 && 
       (depends_S( c->height1) || depends_S( c->height2)))
      parse_error( "requires single height expression for S dependancy");

   if( depends_this_coil( c->height1, c) || 
       depends_this_coil( c->height2, c) ||
       depends_this_coil( c->radius1, c) ||
       depends_this_coil( c->radius2, c) ||
       depends_this_coil( c->wirad, c) ||
       depends_this_coil( c->turns, c)
     )
      parse_error( "self-referencing coil description");
   return 1;
}

/* 
 *  A statement to set the default units.
 */

int parse_units( void)
{
   if( !parse_keyword( "UNITS")) return 0;

   if( parse_keyword( "METRES") ||
       parse_keyword( "METERS") ||
       parse_keyword( "METRIC"))
   {
      unit_str = "m";
      unit_scale = 1.0;
   }
   else
   if( parse_keyword( "INCHES"))
   {
      unit_str = "\"";
      unit_scale = 1/39.37; /*  For use with 'USA' inches. 
                             *  For 'British' inches use 1/39.370147, and
                             *  For 'Canadian' inches use 1/39.370079
                             */
   }
   else
      parse_error( "length unit not recognised");

   return 1;
}

int parse_ambient( void)
{
  if( parse_keyword( "AMBIENT"))
  { 
     ambient = parse_tertiary_expression();
     return 1;
  }
  else return 0;
}

void setup_constants( void)
{
   /*
    *  Our special variable 'S', used as the parameter when specifying non-
    *  uniform coil profiles.
    */

   S = &(alloc_defn( "S")->expr = alloc_node( NT_CONST))->u.val;

   /*
    *  Define a variable called PI.
    */

   (alloc_defn( "PI")->expr = alloc_node( NT_CONST))->u.val = M_PI;
}

   void output_inductance();

int parse_show( void)
{
   struct SHOW *s;

   if( !parse_keyword( "SHOW")) return 0;

   s = alloc_show( parse_tertiary_expression()); 
   while( 1)
   {
      if( parse_keyword( "HEADING") ||
          parse_keyword( "AS")) 
      {
         if( !parse_qstring( s->heading) &&
             !parse_name( s->heading)) parse_error( "missing heading");
      }
      else
      if( parse_keyword( "FORMAT") ||
          parse_keyword( "WITH"))
      {
         if( parse_keyword( "INDUCTANCE")) s->fcn = output_inductance;
         else
         if( !parse_qstring( s->format)) parse_error( "missing format");
      }
      else
         break;
   }

   /*
    *  We insist on a column heading.
    */

   if( !s->heading[0]) parse_error( "missing heading");

   /*
    *  Default the format to 7.3, and check that the format specifier is at 
    *  least d.d, and choose a column width s->len. 
    */

   if( !s->fcn)
   {
      if( !s->format[0]) strcpy( s->format, "7.3");
      if( !isdigit( s->format[0]) ||
          s->format[1] != '.' ||
          !isdigit( s->format[2])) parse_error( "invalid format specifier");
      s->len = s->format[0] - '0' + strlen( s->format+3);
   }
   else s->len = 10;

   if( strlen( s->heading) > s->len) s->len = strlen( s->heading);

   return 1;
}

struct COIL *profile_from = 0;
struct COIL *profile_coil = 0;

int parse_profile( void)
{
   char temp[ 100];

   if( !parse_keyword( "PROFILE")) return 0;

   if( profile_coil)
      parse_error( "cannot have more than one profile statement");
 
   if( !parse_name( temp))
      parse_error( "expecting coil name after PROFILE");

   if( (profile_coil = find_coil( temp)) == 0)
      parse_error( "undefined coil [%s] for profiling", temp);

   if( !parse_keyword( "FROM"))
      parse_error( "expecting FROM in profile");

   if( !parse_name( temp))
      parse_error( "expecting coil name after FROM");

   if( (profile_from = find_coil( temp)) == 0)
      parse_error( "undefined coil [%s] for profiling", temp);
   
   return 1;
}

struct COIL *field_coil = 0;

int parse_field_coil( void)
{
   char temp[ 100];

   if( !parse_keyword( "FIELD")) return 0;

   if( field_coil)
      parse_error( "cannot have more than one field statement");
 
   if( !parse_name( temp))
      parse_error( "expecting coil name after FIELD");

   if( (field_coil = find_coil( temp)) == 0)
      parse_error( "undefined coil [%s] for field plot", temp);

   if( parse_keyword( "OVER"))
   {
      field_maxr = parse_scaled_expression();
      parse_char( ',');
      field_minz = parse_scaled_expression();
      parse_char( ',');
      field_maxz = parse_scaled_expression();
   }
   return 1;
}

/*
 *  Parse a request for optimisation.  For the optimisation to be meaningful,
 *  there must already be at least one range variable specified earlier in the
 *  input file, and the optimised expression must not contain a range variable.
 */

struct NODE *maximise = 0;
struct NODE *minimise = 0;

int parse_optimise_statement( void)
{
   if( parse_keyword( "MAXIMISE"))
   { 
      if( maximise || minimise)
         parse_error( "cannot have more than one optimisation");

      maximise = parse_tertiary_expression();
      if( depends_S( maximise))
         parse_error( "expression cannot involve S");
      if( !fixed( maximise))
         parse_error( "optimisation contains meaningless range");
   }
   else
   if( parse_keyword( "MINIMISE"))
   { 
      if( maximise || minimise)
         parse_error( "cannot have more than one optimisation");

      minimise = parse_tertiary_expression();
      if( depends_S( minimise))
         parse_error( "expression cannot involve S");
      if( !fixed( minimise))
         parse_error( "optimisation contains meaningless range");
   }
   else return 0;

   if( !ranges)
      parse_error( "needs a range of parameters for optimisation");
   return 1;
}

/*
 *  This function is called if the input file doesn't contain any 'show'
 *  statements.  A set of show statements are generated internally to give
 *  the default output columns.
 */

void setup_default_output( void)
{
   struct DEFN *d = defns;
   struct SHOW *s;
   struct NODE *n;
   struct COIL *c, *ci, *cj;

   /*
    *  For each variable definition that includes a range, add a column
    *  to show its value.
    */
 
   for( ; d; d = d->next)
      if( !fixed( d->expr) || opt_ALL)
         if( !depends_S( d->expr))
         {
            (n = alloc_node( NT_DEFN))->u.defn = d;
            s = alloc_show( n); 
            strcpy( s->heading, d->name);
            s->len = strlen( d->name);
            if( s->len < 7) s->len = 7;
            strcpy( s->format, "7.3");
         }

   /* 
    *  For each coil, add a column to show its variable dimensions, if any.
    */

   for( c = coils; c; c = c->next)
   {
      if( c->radius1 == c->radius2)
      {
         if( !fixed( c->radius1) &&
             !depends_S( c->radius1)) 
         {
            s = alloc_show( c->radius1);
            sprintf( s->heading, "%s.r", c->abbr);
            strcpy( s->format, "7.3");
            s->len = 7;
         }
      } 
      else
      {
         if( !fixed( c->radius1))
         {
            s = alloc_show( c->radius1);
            sprintf( s->heading, "%s.r1", c->abbr);
            strcpy( s->format, "7.3");
            s->len = 7;
         }
         if( !fixed( c->radius2))
         {
            s = alloc_show( c->radius2);
            sprintf( s->heading, "%s.r2", c->abbr);
            strcpy( s->format, "7.3");
            s->len = 7;
         }
      }

      if( c->height1 == c->height2)
      {
         if( !fixed( c->height1) &&
             !depends_S( c->height1))
         {
            s = alloc_show( c->height1);
            sprintf( s->heading, "%s.h", c->abbr);
            strcpy( s->format, "7.3");
            s->len = 7;
         } 
      }
      else
      {
         if( !fixed( c->height1)) 
         {
            s = alloc_show( c->height1);
            sprintf( s->heading, "%s.h1", c->abbr);
            strcpy( s->format, "7.3");
            s->len = 7;
         }
         if( !fixed( c->height2))
         { 
            s = alloc_show( c->height2);
            sprintf( s->heading, "%s.h2", c->abbr);
            strcpy( s->format, "7.3");
            s->len = 7;
         }
      }

      if( !fixed( c->turns))
      {
         s = alloc_show( c->turns);
         sprintf( s->heading, "%s.n", c->abbr);
         strcpy( s->format, "7.2");
         s->len = 7;
      }

      if( !fixed( c->wirad)) 
      {
         s = alloc_show( c->wirad);
         sprintf( s->heading, "%s.w", c->abbr);
         strcpy( s->format, "7.3");
         s->len = 7;
      }
   }

   /*
    *  Add columns for each coil's self inductance and resistance.
    */

   if( opt_L)
      for( c = coils; c; c = c->next)
      {
         (n = alloc_node( NT_COIL))->u.coil = c;
         (end_of(n)->next =
                   alloc_node( NT_CFCN1))->u.func = compute_self_inductance;
         s = alloc_show( n);
         sprintf( s->heading, "%s.L", c->abbr);
         s->len = 10;
         s->fcn = output_inductance;
   
         (n = alloc_node( NT_COIL))->u.coil = c;
         (end_of(n)->next =
                   alloc_node( NT_CFCN1))->u.func = compute_resistance;
         s = alloc_show( n);
         sprintf( s->heading, "%s.R", c->abbr);
         s->len = 9;
         strcpy( s->format, "9.5");
      }

   /*
    *  Add mutual inductance and coupling columns for each pair of coils.
    */

   if( opt_M)
      for( ci = coils; ci; ci = ci->next)
         for( cj = ci->next; cj; cj = cj->next)
         {
            (n = alloc_node( NT_COIL))->u.coil = ci;
            (end_of(n)->next = alloc_node( NT_COIL))->u.coil = cj;
            (end_of(n)->next =
                    alloc_node( NT_CFCN2))->u.func = compute_mutual_inductance;
            s = alloc_show( n);
            sprintf( s->heading, "%s-%s.M", ci->abbr, cj->abbr);
            s->len = 11;
            s->fcn = output_inductance;
   
            (n = alloc_node( NT_COIL))->u.coil = ci;
            (end_of(n)->next = alloc_node( NT_COIL))->u.coil = cj;
            (end_of(n)->next =
                    alloc_node( NT_CFCN2))->u.func = compute_coupling_coeff;
            s = alloc_show( n);
            sprintf( s->heading, "%s-%s.K", ci->abbr, cj->abbr);
            s->len = 11;
            strcpy( s->format, "6.4");
         }  
}

/*
 *  Top entry point to the parser.
 */

void parse_input_file( void)
{
   pbuff[ 0] = 0; pb = pbuff; eof_flag = lino = 0;      /* Initialise parser */

   while( skip_whitespace(), !eof_flag)
   {
      char name[ 100];

      if( parse_units() ||
          parse_ambient() ||
          parse_show() ||
          parse_profile() ||
          parse_field_coil() ||
          parse_optimise_statement()) continue;

      if( !parse_name( name) ||
          (!parse_defn( name) && 
             !parse_coil( name))) parse_error( "unrecognised statement");
   }

   if( !coils) parse_error( "no coils defined");
   if( profile_coil && ranges)
      parse_error( "cannot have ranges when profiling");
}

void check_parameters( struct COIL *c)
{
   if( evaluate( c->turns) == 0)
      bailout( "%s cannot have zero turns\n", c->name);
 
   if( evaluate_scaled( c->turns) < 0)
      bailout( "%s cannot have negative turns\n", c->name);
 
   if( evaluate_scaled( c->wirad) == 0)
      bailout( "%s cannot have zero conductor radius\n", c->name);
 
   if( evaluate_scaled( c->wirad) < 0)
      bailout( "%s cannot have negative conductor radius\n", c->name);

#if 0 
   if( evaluate_scaled( c->radius1) == 0 || evaluate( c->radius2) == 0)
      if( !depends_S( c->radius1))
         bailout( "%s cannot have zero radius\n", c->name);
#endif

   if( evaluate_scaled( c->radius1) < 0 || evaluate( c->radius2) < 0)
      if( !depends_S( c->radius1))
         bailout( "%s cannot have negative radius\n", c->name);
}

/*****************************************************************************/
/*  Output Functions                                                         */
/*****************************************************************************/

/*
 *  A couple of state variables to help with placing field and record
 *  delimiters.
 */

int ofs_flag = 0;            /* Zero if an output field separator is pending */
int inside_header = 0;          /* Non-zero while we're outputing the header */

void begin_output( void)
{
   if( opt_HTML)
   {
      printf( "<table border=1>\n");
      return;
   }
}

void end_output( void)
{
   if( opt_HTML)
   {
      printf( "</table><p>\n");
      return;
   }
}

void begin_header( void)
{
   inside_header = 1;

   if( opt_HTML)
   {
      printf( "<tr>\n");
      return;
   }
}

void end_header( void)
{
   inside_header = 0;

   if( opt_HTML)
   {
      printf( "</tr>\n");
      return;
   }
}

void begin_row( void)
{
   if( opt_HTML)
   {
      printf( "<tr>\n");
      return;
   }
   ofs_flag = 1;
}

void end_row( void)
{
   if( opt_HTML)
   {
      printf( "</tr>\n");
      return;
   }
   printf( "\n");
}

void ofs( void)
{
   if( opt_HTML)
   {
      return;
   }
   if( opt_CSV)
   {
      printf( ",");
      return;
   }
   if( opt_SSV)
   {
      printf( " ");
      return;
   }
   printf( "|");
}

void output_field( int len, char *format, ...)
{
   char temp[100];

   if( !ofs_flag) ofs();
   ofs_flag = 0;

   vsprintf( temp, format, &format + 1);

   if( opt_HTML)
   {
      if( inside_header) printf( "<th align=center>%s</th>", temp);
      else printf( "<td align=right>%s</td>", temp);
      return;
   }

   if( opt_CSV || opt_SSV)
   {
      printf( "%s", temp);
      return;
   }

   /*
    *  Fall through to plain text output, pad field to required length.
    */

   printf( "%*s", len, temp);
}

void xx_output_field( struct SHOW *s)
{
   char format[50];
   double val = evaluate( s->expr);

   if( !s->fcn)
   {
      sprintf( format, "%%%3.3sf%s", s->format, s->format+3);
      output_field( s->len, format, val);
   }
   else
   {
      (s->fcn)( s->len, val);
      return;
   }
}

void output_inductance( int len, double h)
{
   if( opt_CSV || opt_SSV)
   {
      output_field( len, "%.4e", h);
      return;
   }

   if( h < 10e-3) output_field( len, "%.2f uH", h * 1e6);
   else
   if( h < 1000e-3) output_field( len, "%.2f mH", h * 1e3);
   else output_field( len, "%.3f H", h);
}

/*****************************************************************************/
/*  Inductance and Resistance.                                               */
/*****************************************************************************/

/*
 *  At last...we get to some physics.
 */

#define ND 300        /* Each winding is decomposed into this many elements */

/*
 *  Calculate the DC resistance of the winding. 1.75e-8 is the resistivity of
 *  copper at 25 deg C.
 */

double compute_length( struct COIL *c)
{
   int j;
   int sr = depends_S( c->radius1);
   double R1 = evaluate_scaled( c->radius1);
   double R2 = evaluate_scaled( c->radius2);
   double dR = (R2 - R1)/ND;
   double length = 0;

   for( j=0; j<ND; j++)
      if( sr)
      {
         *S = (j+0.5)/ND;
         length += evaluate_scaled( c->radius1);
      }
      else
          length += (j+0.5) * dR + R1;

   return length * 2 * M_PI * evaluate( c->turns)/ND;
}

double compute_resistance( struct COIL *c)
{
   double w = evaluate_scaled( c->wirad);
   double cu;
   double a = (ambient != NULL) ? evaluate( ambient) : 25.0;

   cu = (a/25.0 * (1.75 - 1.58) + 1.58) * 1e-8;
   return compute_length( c)/(M_PI * w * w) * cu;
}

/*
 *  Compute the self inductance of winding c.
 */

double compute_self_inductance( struct COIL *c)
{
   int j, k;
   int sr = depends_S( c->radius1);
   int sh = depends_S( c->height1);
   int sI = depends_S( c->current);
   double L = 0;
   double R1 = evaluate_scaled( c->radius1);
   double R2 = evaluate_scaled( c->radius2);
   double H1 = evaluate_scaled( c->height1);
   double H2 = evaluate_scaled( c->height2);
   double t = evaluate( c->turns);
   double dH = (H2 - H1)/ND;
   double dR = (R2 - R1)/ND;
   double w = evaluate_scaled( c->wirad);

   for( j=0; j<ND; j++)
   {
      for( k=0; k<ND; k++)
      {
         double r1, r2, d;

         if( sh)
         {
            *S = (j+0.5)/ND;
            d = evaluate_scaled( c->height1);
            *S = (k+0.5)/ND;
            d -= evaluate_scaled( c->height1);
         }
         else d = (j - k) * dH;

         if( sr)
         {
            *S = (j+0.5)/ND;
            r1 = evaluate_scaled( c->radius1);
            *S = (k+0.5)/ND;
            r2 = evaluate_scaled( c->radius1);
         }
         else
         {
            r1 = (j + 0.5) * dR + R1;
            r2 = (k + 0.5) * dR + R1;
         }  

         if( sI)
         {
            *S = (k+0.5)/ND;
            L += filament( r1, r2, d, w) * evaluate( c->current);
         }
         else
            L += filament( r1, r2, d, w);
      }
   }

   if( sI)
   {
      double end_1, end_2;

      *S = 0; end_1 = evaluate( c->current);
      *S = 1; end_2 = evaluate( c->current);

      L /= end_1 > end_2 ? end_1 : end_2;
   }

   return L * t*t/(ND*(double)ND);
}

/*
 *  Compute the mutual inductance between windings ca and cb.
 */

double compute_mutual_inductance( struct COIL *ca, struct COIL *cb)
{
   int j, k;
   int asr = depends_S( ca->radius1);
   int ash = depends_S( ca->height1);
   int bsr = depends_S( cb->radius1);
   int bsh = depends_S( cb->height1);
   int sI = depends_S( cb->current);
   double M = 0;
   double aR1 = evaluate_scaled( ca->radius1);
   double aR2 = evaluate_scaled( ca->radius2);
   double aH1 = evaluate_scaled( ca->height1);
   double aH2 = evaluate_scaled( ca->height2);
   double bR1 = evaluate_scaled( cb->radius1);
   double bR2 = evaluate_scaled( cb->radius2);
   double bH1 = evaluate_scaled( cb->height1);
   double bH2 = evaluate_scaled( cb->height2);
   double aw = evaluate_scaled( ca->wirad);
   double bw = evaluate_scaled( cb->wirad);
   double adH = (aH2 - aH1)/ND;
   double adR = (aR2 - aR1)/ND;
   double bdH = (bH2 - bH1)/ND;
   double bdR = (bR2 - bR1)/ND;

   for( j=0; j<ND; j++)
      for( k=0; k<ND; k++)
      {
         double r1, r2, d;

         if( ash)
         {
            *S = (j+0.5)/ND;
            d = evaluate_scaled( ca->height1);
         }
         else d = aH1 + (j + 0.5) * adH;

         if( bsh)
         {
            *S = (k+0.5)/ND;
            d -= evaluate_scaled( cb->height1);
         }
         else d -= bH1 + (k + 0.5) * bdH;

         if( asr)
         {
            *S = (j+0.5)/ND;
            r1 = evaluate_scaled( ca->radius1);
         }
         else r1 = (j + 0.5) * adR + aR1 - aw;         

         if( bsr)
         {
            *S = (k+0.5)/ND;
            r2 = evaluate_scaled( cb->radius1);
         }
         else r2 = (k + 0.5) * bdR + bR1 - bw;         

         if( sI)
         {
            *S = (k+0.5)/ND;
            M += filament( r1, r2, d, (aw+bw)/2) * evaluate( cb->current);
         }
         else
            M += filament( r1, r2, d, (aw+bw)/2);
      }

   if( sI)
   {
      double end_1, end_2;

      *S = 0; end_1 = evaluate( cb->current);
      *S = 1; end_2 = evaluate( cb->current);

      M /= end_1 > end_2 ? end_1 : end_2;
   }

   return M * evaluate( ca->turns) * evaluate( cb->turns)/(ND*(double)ND);
}

double compute_coupling_coeff( struct COIL *ca, struct COIL *cb)
{
   return compute_mutual_inductance( ca, cb) / 
          sqrt( compute_self_inductance( ca) * compute_self_inductance( cb));
}

/*
 *  Profile the induced voltage in coil ca given a dI/dt in coil cb of 1A/sec.
 */

void profile_induction( struct COIL *ca, struct COIL *cb)
{
   int j, k;
   int asr = depends_S( ca->radius1);
   int ash = depends_S( ca->height1);
   int bsr = depends_S( cb->radius1);
   int bsh = depends_S( cb->height1);
   double aR1 = evaluate_scaled( ca->radius1);
   double aR2 = evaluate_scaled( ca->radius2);
   double aH1 = evaluate_scaled( ca->height1);
   double aH2 = evaluate_scaled( ca->height2);
   double bR1 = evaluate_scaled( cb->radius1);
   double bR2 = evaluate_scaled( cb->radius2);
   double bH1 = evaluate_scaled( cb->height1);
   double bH2 = evaluate_scaled( cb->height2);
   double aw = evaluate_scaled( ca->wirad);
   double bw = evaluate_scaled( cb->wirad);
   int Na = (int)(0.5 + evaluate( ca->turns)); 
   double adH = (aH2 - aH1)/Na;
   double adR = (aR2 - aR1)/Na;
   double bdH = (bH2 - bH1)/ND;
   double bdR = (bR2 - bR1)/ND;
   double V = 0;
   double dV;

   for( j=0; j<Na; j++)
   {
      dV = 0;

      for( k=0; k<ND; k++)
      {
         double r1, r2, d;

         if( ash)
         {
            *S = (j+0.5)/Na;
            d = evaluate_scaled( ca->height1);
         }
         else d = aH1 + (j + 0.5) * adH;

         if( bsh)
         {
            *S = (k+0.5)/ND;
            d -= evaluate_scaled( cb->height1);
         }
         else d -= bH1 + (k + 0.5) * bdH;

         if( asr)
         {
            *S = (j+0.5)/Na;
            r1 = evaluate_scaled( ca->radius1);
         }
         else r1 = (j + 0.5) * adR + aR1 - aw;         

         if( bsr)
         {
            *S = (k+0.5)/ND;
            r2 = evaluate_scaled( cb->radius1);
         }
         else r2 = (k + 0.5) * bdR + bR1 - bw;         

         dV += filament( r1, r2, d, (aw+bw)/2);
      }

      dV *= evaluate( cb->turns)/((double)ND);
      V += dV;
      begin_row();
      output_field( 0, "%d", j+1);
      output_field( 0, "%.3e", dV); 
      output_field( 0, "%.3e", V); 
      end_row();
   }
}

/*****************************************************************************/
/*  Main                                                                     */
/*****************************************************************************/

void generate_results( void)
{
   struct SHOW *s = shows;

   begin_row();
      for( ; s; s = s->next) xx_output_field( s);
   end_row();
}

double maxval = -1e99;
double minval = 1e99;

void maximise_results( void)
{
   double v = evaluate( maximise);

   if( v > maxval) 
   {
      maxval = v;
      save_optimum();
   }
}

void minimise_results( void)
{
   double v = evaluate( minimise);

   if( v < minval) 
   {
      minval = v;
      save_optimum();
   }
}

/*
 *  Sweep through all values of the range variables, calling the generator
 *  function for each permutation.
 */

void sweep_ranges( struct RANGE *r, void (*generator)(void))
{
   struct COIL *c;

   /*
    *  Use a recursion to do a depth-first descent down the linked list of
    *  range variables in order to visit every possible permutation.
    */

   if( r)
   {
      for( r->val = evaluate( r->from);
           r->val * PREC <= evaluate( r->to);
           r->val += evaluate( r->step)) sweep_ranges( r->next, generator);
      return;
   } 

   /*
    *  We emerge here for each permutation of range variable values.  Make sure
    *  the permutation is physically meaningful.
    */

   for( c = coils; c; c = c->next) check_parameters( c);

   /*
    *  Everything is setup. Now call the generator function to output a row
    *  of data.
    */

   *S = 0;
   generator();
}

#define FIELD_STEPS 25

#define FF_LMARG 50
#define FF_TMARG 50
#define FF_HEIGHT 612
#define FF_WIDTH 792
#if 0
#define SCALEY(A) \
  (( FF_HEIGHT-((A) * (FF_HEIGHT - 2*FF_LMARG) + FF_LMARG))/0.06)
#define SCALEX(A) \
  (((A) * (FF_WIDTH - 2*FF_TMARG)/2)/0.06)

void draw_line( int col, double x1, double y1, double x2, double y2)
{
   printf( "n %d %d m %d %d l gs col%d s gr\n",
         (int)SCALEY(y1), (int)(SCALEX(x1) + FF_WIDTH/2/0.06), 
         (int)SCALEY(y2), (int)(SCALEX(x2) + FF_WIDTH/2/0.06), col);
   printf( "n %d %d m %d %d l gs col%d s gr\n",
         (int)SCALEY(y1), (int)(-SCALEX(x1) + FF_WIDTH/2/0.06), 
         (int)SCALEY(y2), (int)(-SCALEX(x2) + FF_WIDTH/2/0.06), col);
}
#endif

void xdraw_line( int col, double r1, double z1, double r2, double z2)
{
   r1 = (r1 - fminr)/(fmaxr-fminr) * (FF_WIDTH-2*FF_LMARG)/2;
   r2 = (r2 - fminr)/(fmaxr-fminr) * (FF_WIDTH-2*FF_LMARG)/2;
   z1 = (z1 - fminz)/(fmaxz-fminz) * (FF_HEIGHT-2*FF_TMARG) + FF_TMARG;
   z2 = (z2 - fminz)/(fmaxz-fminz) * (FF_HEIGHT-2*FF_TMARG) + FF_TMARG;
   printf( "n %d %d m %d %d l gs col%d s gr\n",
         (int)((FF_HEIGHT-z1)/0.06), (int)((FF_WIDTH/2 - r1)/0.06), 
         (int)((FF_HEIGHT-z2)/0.06), (int)((FF_WIDTH/2 - r2)/0.06), col);
   printf( "n %d %d m %d %d l gs col%d s gr\n",
         (int)((FF_HEIGHT-z1)/0.06), (int)((FF_WIDTH/2 + r1)/0.06), 
         (int)((FF_HEIGHT-z2)/0.06), (int)((FF_WIDTH/2 + r2)/0.06), col);
}

void xdraw_dot( int col, double r1, double z1)
{
   int r,z, k=20;
   r1 = (r1 - fminr)/(fmaxr-fminr) * (FF_WIDTH-2*FF_LMARG)/2;
   z1 = (z1 - fminz)/(fmaxz-fminz) * (FF_HEIGHT-2*FF_TMARG) + FF_TMARG;

   r = (int)((FF_HEIGHT-z1)/0.06);
   z = (int)((FF_WIDTH/2 - r1)/0.06);

   printf( "n %d %d m %d %d l %d %d l %d %d l cp gs col%d s gr\n",
         (int)((FF_HEIGHT-z1)/0.06)-k, (int)((FF_WIDTH/2 - r1)/0.06)-k, 
         (int)((FF_HEIGHT-z1)/0.06)-k, (int)((FF_WIDTH/2 - r1)/0.06)+k, 
         (int)((FF_HEIGHT-z1)/0.06)+k, (int)((FF_WIDTH/2 - r1)/0.06)+k, 
         (int)((FF_HEIGHT-z1)/0.06)+k, (int)((FF_WIDTH/2 - r1)/0.06)-k, 
          col);
   printf( "n %d %d m %d %d l %d %d l %d %d l cp gs col%d s gr\n",
         (int)((FF_HEIGHT-z1)/0.06)-k, (int)((FF_WIDTH/2 + r1)/0.06)-k, 
         (int)((FF_HEIGHT-z1)/0.06)-k, (int)((FF_WIDTH/2 + r1)/0.06)+k, 
         (int)((FF_HEIGHT-z1)/0.06)+k, (int)((FF_WIDTH/2 + r1)/0.06)+k, 
         (int)((FF_HEIGHT-z1)/0.06)+k, (int)((FF_WIDTH/2 + r1)/0.06)-k, 
          col);
}

void draw_coils( double maxd)
{
   struct COIL *c = coils;
#define NC 20
   for( ; c; c = c->next)
   {
      int j;
      int sr = depends_S( c->radius1);
      int sh = depends_S( c->height1);
      double R1 = evaluate_scaled( c->radius1);
      double R2 = evaluate_scaled( c->radius2);
      double H1 = evaluate_scaled( c->height1);
      double H2 = evaluate_scaled( c->height2);
      double dH = (H2 - H1)/NC;
      double dR = (R2 - R1)/NC;

      for( j=0; j<NC; j++)
      {
         double r1, z1, r2, z2;

         if( sh)
         {
            *S = (j+0)/NC;
            z1 = evaluate_scaled( c->height1);
            *S = (j+1)/NC;
            z2 = evaluate_scaled( c->height1);
         }
         else
         {
            z1 = (j+0) * dH + H1;
            z2 = (j+1) * dH + H1;
         }

         if( sr)
         {
            *S = (j+0)/NC;
            r1 = evaluate_scaled( c->radius1);
            *S = (j+1)/NC;
            r2 = evaluate_scaled( c->radius1);
         }
         else
         {
            r1 = (j + 0) * dR + R1;
            r2 = (j + 1) * dR + R1;
         }
         
         xdraw_line( 3, r1, z1, r2, z2);
      }
   } 
}

void set_plot_size( void)
{
   struct COIL *c = coils;
   double t, d;

   fminr = 0;

   if( field_maxr)
   {
      fmaxr = evaluate_scaled( field_maxr);
      fmaxz = evaluate_scaled( field_maxz);
      fminz = evaluate_scaled( field_minz);
      return;
   }

   for( ; c; c = c->next)
   {
      *S = 0;
      t = evaluate_scaled( c->radius1); 
      if( t > fmaxr) fmaxr = t;  
      t = evaluate_scaled( c->height1); 
      if( t > fmaxz) fmaxz = t;  
      if( t < fminz) fminz = t; 

      *S = 1; 
      t = evaluate_scaled( c->radius2); 
      if( t > fmaxr) fmaxr = t;  
      t = evaluate_scaled( c->height2); 
      if( t > fmaxz) fmaxz = t;  
      if( t < fminz) fminz = t; 
   }

   fmaxr *= 2.5;
   d = (fmaxz - fminz)*1.5; 
   t = (fmaxz + fminz)/2;
   if( d == 0) { fmaxz = t + fmaxr; fminz = t - fmaxr; }
   else
   {
      fmaxz = t + d/2;
      fminz = t - d/2;
   }
}

void set_line_width( double t)
{
   printf( "%.3f slw\n", t);
}

void output_field_grid( struct COIL *c, 
                        double *g, int ng)
{
   int ir, iz;
   double maxd = fmaxr > fmaxz ? fmaxr : fmaxz;
   double *gp, norm;

   printf( "%%!PS-Adobe-2.0\n%%%%Title: %s.ps\n", c->name);
   printf( "%%%%Creator: acmi %s\n", VERSION);
   printf( "%%%%Orientation: Landscape\n%%%%Pages: 1\n");
   printf( "%%%%BoundingBox: 0 0 612 792\n%%%%BeginSetup\n");
   printf( "%%%%IncludeFeature: *PageSize Letter\n%%%%EndSetup\n");
   printf( "%%%%Magnification: 1.0000\n%%%%EndComments\n");
   printf( "/$F2psDict 200 dict def\n$F2psDict begin\n");
   printf( "$F2psDict /mtrx matrix put\n");
   printf( "/col-1 {0 setgray} bind def\n");
   printf( "/col0 {0.000 0.000 0.000 srgb} bind def\n");
   printf( "/col1 {0.000 0.000 1.000 srgb} bind def\n");
   printf( "/col2 {0.000 1.000 0.000 srgb} bind def\n");
   printf( "/col3 {0.500 0.000 0.000 srgb} bind def\n");
   printf( "end\nsave\n");

   printf( "newpath %d %d moveto %d %d lineto %d %d lineto %d %d lineto ",
              FF_TMARG, FF_WIDTH-FF_LMARG,
              FF_TMARG, FF_LMARG,
              FF_HEIGHT-FF_TMARG, FF_LMARG,
              FF_HEIGHT-FF_TMARG, FF_WIDTH-FF_LMARG);
   printf( "closepath clip\n");

   printf( "/gr {grestore} bind def\n/gs {gsave} bind def\n");
   printf( "/rs {restore} bind def\n/l {lineto} bind def\n");
   printf( "/m {moveto} bind def\n/n {newpath} bind def\n");
   printf( "/s {stroke} bind def\n/slw {setlinewidth} bind def\n");
   printf( "/cp {closepath} bind def\n");
   printf( "/srgb {setrgbcolor} bind def\n/sc {scale} bind def\n");
   printf( "/$F2psBegin {$F2psDict begin /$F2psEnteredState save def} def\n");
   printf( "/$F2psEnd {$F2psEnteredState restore end} def\n");

   printf( "$F2psBegin\n%%%%Page: 1 1\n10 setmiterlimit 0.06000 0.06000 sc\n");
   printf( "%%%%Page: 1 1\n");

   set_line_width( 7.5);

   xdraw_line( 1, fminr, fmaxz, fmaxr, fmaxz);
   xdraw_line( 1, fmaxr, fmaxz, fmaxr, fminz);
   xdraw_line( 1, fmaxr, fminz, fminr, fminz);

   for( gp = g, norm=0, ir=0; ir<ng; ir++) 
      for( iz=0; iz<ng; iz++) norm += *gp++;
   norm /= ng * ng;

   for( gp = g, ir=0; ir<ng; ir++)
      for( iz=0; iz<ng; iz++)
      {
         double Bx = *gp++;
         double Bz = *gp++;
         double r = fminr + ir * (fmaxr-fminr)/FIELD_STEPS;
         double z = fminz + iz * (fmaxz-fminz)/FIELD_STEPS;
         double mag = maxd/200/norm;

/*         printf( "%.3e %.3e   %.3e %.3e %.3e\n", r, z, Bx, By, Bz); */

         xdraw_dot( 0, r, z);
         xdraw_line( 0, r-mag*Bx/2, z-mag*Bz/2, r+mag*Bx/2, z+mag*Bz/2);
         // xdraw_line( 0, r, z, r+mag*Bx, z+mag*Bz);
      }

   set_line_width( 40.0);
   draw_coils( maxd);

   printf( "newpath %d %d moveto %d %d lineto %d %d lineto %d %d lineto ",
              0, FF_WIDTH,
              0, 0,
              FF_HEIGHT, 0,
              FF_HEIGHT, FF_WIDTH);
   printf( "closepath clip\n");

   printf( "$F2psEnd\nrs\nshowpage\n");

}

void field_plot( struct COIL *c)
{
   int j, ir, iz;
   int sr = depends_S( c->radius1);
   int sh = depends_S( c->height1);
   int sI = depends_S( c->current);
   double R1 = evaluate_scaled( c->radius1);
   double R2 = evaluate_scaled( c->radius2);
   double H1 = evaluate_scaled( c->height1);
   double H2 = evaluate_scaled( c->height2);
   int turns = evaluate( c->turns);
   int NF = turns * 500;
   double dH = (H2 - H1)/NF;
   double dR = (R2 - R1)/NF;
   double *g, *gp;
   int ng;

   set_plot_size();

   ng = FIELD_STEPS;
   g = safe_malloc( 2 * ng * ng * sizeof( double));
   for( j = 0; j< 2 * ng; j++) g[j] = 0;

   for( j=0; j<NF; j++)
   {
      double I, h, r1;

      if( sh)
      {
         *S = (j+0.5)/NF;
         h = evaluate_scaled( c->height1);
      }
      else
         h = (j + 0.5) * dH + H1;

      if( sr)
      {
         *S = (j+0.5)/NF;
         r1 = evaluate_scaled( c->radius1);
      }
      else
         r1 = (j + 0.5) * dR + R1;

      if( sI)
      {
         *S = (j+0.5)/NF;
         I = evaluate( c->current);
      }
      else
         I = 1.0;

      for( gp = g, ir=0; ir<ng; ir++)
         for( iz=0; iz<ng; iz++)
         {
            double d;
            double r = fminr + ir * (fmaxr-fminr)/FIELD_STEPS;
            double z = fminz + iz * (fmaxz-fminz)/FIELD_STEPS;
            double wx, wy, wz, rx, ry, rz, cx, cy, cz;
            double bx, bz;
            double theta = 2 * M_PI * 
                           ((j+0.5) * turns)/NF - (int)(((j+0.5)*turns)/NF);
            double cos_theta = cos( theta);
            double sin_theta = sin( theta);

            /* The point on the wire */
            wx = r1 * cos_theta; wy = r1 * sin_theta; wz = h;

            /* vector from the wire point to the summing point (r,z) */
            rx = r - wx; ry = 0 - wy;  rz = z - wz;

            /* The current vector dl */
            cx = -sin_theta; cy = cos_theta; cz = 0;

            /* The cross product */
            bx = cy*rz - ry*cz;  bz = cx*ry - rx*cy;
            d = sqrt( rx*rx + ry*ry + rz*rz);
            if( d/wx < 0.05) continue;
            d = MU * I/(4*M_PI*d*d*d); 

            *gp++ += bx * d;
            *gp++ += bz * d;
         }
   }

   output_field_grid( c, g, ng);

#if 0

   for( ir=0; ir<ng; ir++)
      for( iz=0; iz<ng; iz++)
      {
         double Bx = *(g + (iz*ng + ir)*3 + 0);
         double By = *(g + (iz*ng + ir)*3 + 1);
         double Bz = *(g + (iz*ng + ir)*3 + 2); 
         double r = ir * maxr/FIELD_STEPS;
         double z = iz * maxz/FIELD_STEPS;

         printf( "%.3e %.3e   %.3e %.3e %.3e\n", r, z, Bx, By, Bz);
      }
#endif
}

void usage( void)
{
   fprintf( stderr, "ACMI version %s\n\n", VERSION);

   fprintf( stderr, "usage:    ./acmi [options] < input.txt\n\n");
  
   fprintf( stderr, "options are:\n");
   fprintf( stderr, "  -h         Show this message\n"); 
   fprintf( stderr, "  -heading   Don't print column headings\n"); 
   fprintf( stderr, "  -html      Generate HTML output\n");
   fprintf( stderr, "  -csv       Generate CSV output\n");
   fprintf( stderr, "  -ssv       Generate SSV output\n");
   fprintf( stderr, "  -M         Don't bother with mutual inductances\n");
   fprintf( stderr, "  -L         Don't bother with self inductances\n");
   fprintf( stderr, "  -all       Include all input dimensions in output\n");

   fprintf( stderr, 
      "\nSee http://www.abelian.demon.co.uk/acmi/ for documentation.\n");
   fprintf( stderr, 
      "Maintained by Paul Nicholson, paul@abelian.demon.co.uk\n");
}

int main( int argc, char *argv[])
{
   int i;
   struct SHOW *s;

   /*
    *  Sort out the command line options.
    */

   for( i=1; i<argc; i++)
   {
      if( !strcmp( argv[i], "-heading")) opt_HEADING = 0;
      else
      if( !strcmp( argv[i], "-html")) opt_HTML = 1;
      else
      if( !strcmp( argv[i], "-csv")) opt_CSV = 1;
      else
      if( !strcmp( argv[i], "-ssv")) opt_SSV = 1;
      else
      if( !strcmp( argv[i], "-M")) opt_M = 0;
      else
      if( !strcmp( argv[i], "-L")) opt_L = 0;
      else 
      if( !strcmp( argv[i], "-all")) opt_ALL = 1;
      else 
      if( !strcmp( argv[i], "-h") ||
          !strcmp( argv[i], "--help") ||
          !strcmp( argv[i], "-help")) opt_HELP = 1; 
      else
      {
         usage();
         bailout( "unrecognised option: %s\n", argv[ i]);
      }
   }

   if( opt_HELP) { usage(); exit( 0); }

   setup_constants();
   parse_input_file();

   if( field_coil)
   {
      field_plot( field_coil); 
      return 0;
   }

   if( profile_coil)
   {
      begin_output();
      if( opt_HEADING)
      {
         begin_header();
         begin_row();
              
         output_field( 0, "turn");
         output_field( 0, "dV");
         output_field( 0, "V");

         end_row();
         end_header();
      }
   
      profile_induction( profile_coil, profile_from);
      end_output();
      return 0;
   }

   if( !shows) setup_default_output();

   begin_output();

   if( opt_HEADING)
   {
      begin_header();
      begin_row();
           
      for( s = shows; s; s = s->next) output_field( s->len, "%s", s->heading);
      end_row();
      end_header();
   }

   if( maximise)
   {
      sweep_ranges( ranges, maximise_results);
      restore_optimum();
      begin_row();
         for( s = shows; s; s = s->next) xx_output_field( s);
      end_row();
   }
   else
   if( minimise)
   {
      sweep_ranges( ranges, minimise_results);
      restore_optimum();
      begin_row();
         for( s = shows; s; s = s->next) xx_output_field( s);
      end_row();
   }
   else
      sweep_ranges( ranges, generate_results);

   end_output();

   return 0;
}

